"""
The main IPv6 DHCPd
"""

import argparse
import codecs
import concurrent.futures
import configparser
import importlib
from ipaddress import IPv6Address, AddressValueError
from logging import StreamHandler, Formatter
from logging.handlers import SysLogHandler
import selectors
import socket
from struct import pack
import sys
import os
import re
import signal
import fcntl
import time
import pwd
import grp
import netifaces
import logging
import logging.handlers
import types

import dhcpkit
from dhcpkit.ipv6.duids import DUID, LinkLayerDUID
from dhcpkit.ipv6.exceptions import InvalidPacketError, ListeningSocketError
from dhcpkit.ipv6.message_handlers import MessageHandler
from dhcpkit.ipv6.listening_socket import ListeningSocket
from dhcpkit.ipv6.messages import RelayReplyMessage
from dhcpkit.utils import camelcase_to_dash

logger = logging.getLogger()


class ServerConfigParser(configparser.ConfigParser):
    """
    Special config parser that normalises section names
    """

    class SectionNameNormalisingRegEx:
        """
        Fake regex that normalises its output
        """
        SECTCRE = configparser.ConfigParser.SECTCRE

        def match(self, value: str):
            """
            Fake regex match function that normalises the result and then creates a real match object.

            :param value: the value to match against
            :returns: A match object or None
            """
            # Do matching using the normal re
            matches = self.SECTCRE.match(value)

            # No match: don't change anything
            if not matches:
                return matches

            # Match! Now monkey-patch the result
            header = matches.group('header')
            header = ServerConfigParser.normalise_section_name(header)

            # And recreate
            return self.SECTCRE.match('[{}]'.format(header))

    SECTCRE = SectionNameNormalisingRegEx()

    def optionxform(self, optionstr: str) -> str:
        """
        Transform option names to a standard form. Allow options with underscores and convert those to dashes.

        :param optionstr: The original option name
        :returns: The normalised option name
        """
        return optionstr.lower().replace('_', '-')

    @staticmethod
    def normalise_section_name(section: str) -> str:
        """
        Normalise a section name.

        :param section: The raw name of the section
        :returns: The normalised name
        """
        # Collapse multiple spaces
        section = re.sub(r'[\t ]+', ' ', section)

        # Split
        parts = section.split(' ')
        parts[0] = parts[0].lower()

        # Special section names
        if parts[0] == 'interface':
            # Check name structure
            if len(parts) != 2:
                raise configparser.ParsingError("Interface sections must be named [interface xyz] "
                                                "where 'xyz' is an interface name")

        elif parts[0] == 'option':
            # Check name structure
            if not (2 <= len(parts) <= 3):
                raise configparser.ParsingError("Option sections must be named [option xyz] or [option xyz id]"
                                                "where 'xyz' is an option handler name and 'id' is an identifier "
                                                "to distinguish between multiple handlers of the same type")

            if '-' in parts[1] or '_' in parts[1]:
                parts[1] = parts[1].replace('_', '-').lower()
            else:
                parts[1] = camelcase_to_dash(parts[1])

            # If the name ends with
            if parts[1].endswith('-option-handler'):
                parts[1] = parts[1][:-15]

        elif parts[0] not in ('logging', 'server',):
            raise configparser.ParsingError("Invalid section name: [{}]".format(section))

        # Reconstruct
        return ' '.join(parts)

    def add_section(self, section):
        """
        Also normalise section names that are added by the code.

        :param section: The section name
        """
        section = self.normalise_section_name(section)
        super().add_section(section)


def handle_args():
    """
    Handle the command line arguments.

    :return: The arguments object
    """
    parser = argparse.ArgumentParser(
        description="A flexible IPv6 DHCP server written in Python.",
    )

    parser.add_argument("config", help="the configuration file")
    parser.add_argument("-C", "--show-config", action="store_true", help="show the active configuration")
    parser.add_argument("-v", "--verbosity", action="count", default=0, help="increase output verbosity")

    args = parser.parse_args()

    return args


def load_config(config_filename: str) -> configparser.ConfigParser:
    """
    Load the given configuration file.

    :param config_filename: The configuration file
    :return: The parsed config
    """
    logger.debug("Loading configuration file {}".format(config_filename))

    config = ServerConfigParser()

    # Create mandatory sections and options
    config.add_section('logging')
    config['logging']['facility'] = 'daemon'

    config.add_section('server')
    config['server']['duid'] = 'auto'
    config['server']['message-handler'] = 'standard'
    config['server']['user'] = 'nobody'
    config['server']['group'] = ''
    config['server']['exception-window'] = '1.0'
    config['server']['max-exceptions'] = '10'
    config['server']['threads'] = '10'

    try:
        config_filename = os.path.realpath(config_filename)
        config_dir = os.path.dirname(config_filename)
        config_file = open(config_filename, mode='r', encoding='utf-8')
        config.read_file(config_file)
    except FileNotFoundError:
        logger.error("Configuration file {} not found".format(config_filename))
        sys.exit(1)

    # Assume that every option name that contains the word 'file' or 'filename' is a filename and needs to be
    # converted into an absolute path if it isn't already
    for section in config.sections():
        for option in config[section]:
            if re.match(r'.*\bfile(name)?\b.*', option):
                filename = config[section][option]

                # Make the filename absolute
                if not os.path.isabs(filename):
                    config[section][option] = os.path.realpath(os.path.join(config_dir, filename))

    return config


def set_up_logger(config: configparser.ConfigParser, verbosity: int=0):
    """
    Set up logging based on the information in the configuration.

    :param config: The configuration
    :param verbosity: The verbosity level given as command line argument
    """
    # Don't filter on level in the root logger
    logger.setLevel(logging.NOTSET)

    # Determine syslog facility
    facility_name = config['logging']['facility'].lower()
    facility = logging.handlers.SysLogHandler.facility_names.get(facility_name)
    if not facility:
        logger.critical("Invalid logging facility: {}".format(facility_name))
        sys.exit(1)

    # Create the syslog handler
    syslog_handler = SysLogHandler(facility=facility)
    logger.addHandler(syslog_handler)

    # Also output to sys.stdout
    stdout_handler = StreamHandler(stream=sys.stdout)

    # Set level according to verbosity
    if verbosity >= 3:
        stdout_handler.setLevel(logging.DEBUG)
    elif verbosity == 2:
        stdout_handler.setLevel(logging.INFO)
    elif verbosity >= 1:
        stdout_handler.setLevel(logging.WARNING)
    else:
        stdout_handler.setLevel(logging.CRITICAL)

    # Try using colourised output
    try:
        # noinspection PyPackageRequirements
        import colorlog
    except ImportError:
        colorlog = None

    if colorlog:
        if verbosity >= 3:
            # The color names are black, red, green, yellow, blue, purple, cyan and white.
            formatter = colorlog.ColoredFormatter('{yellow}{asctime}{reset} '
                                                  '[{threadName}] '
                                                  '{cyan}{name}#{lineno}{reset} '
                                                  '[{log_color}{levelname}{reset}] '
                                                  '{message}', style='{')
        elif verbosity == 2:
            formatter = colorlog.ColoredFormatter('{yellow}{asctime}{reset} '
                                                  '[{log_color}{levelname}{reset}] '
                                                  '{message}', datefmt=Formatter.default_time_format, style='{')
        else:
            formatter = None

    else:
        # Set output style according to verbosity
        if verbosity >= 3:
            formatter = Formatter('{asctime} [{threadName}] {name}#{lineno} [{levelname}] {message}', style='{')
        elif verbosity == 2:
            formatter = Formatter('{asctime} [{levelname}] {message}', datefmt=Formatter.default_time_format,
                                  style='{')
        else:
            formatter = None

    stdout_handler.setFormatter(formatter)
    logger.addHandler(stdout_handler)


def get_handler(config: configparser.ConfigParser) -> MessageHandler:
    """
    Get the request handler specified in the configuration file.

    :param config: The configuration
    :return: The handler
    """
    handler_module_name = config['server'].get('message-handler')

    if not handler_module_name:
        logger.critical("No message handler configured")
        sys.exit(1)

    try:
        logger.debug("Trying to import message handler from {}".format(handler_module_name))
        handler_module = importlib.import_module(handler_module_name)
        logger.info("Imported message handler from {}".format(handler_module_name))
    except ImportError as e:
        if '.' in handler_module_name:
            # It was a qualified name, and it didn't work...
            logger.critical(str(e))
            sys.exit(1)

        # No dots, try to prepend the default module path
        try:
            handler_module_name = 'dhcpkit.ipv6.message_handlers.' + handler_module_name

            logger.debug("Trying to import message handler from {}".format(handler_module_name))
            handler_module = importlib.import_module(handler_module_name)
            logger.info("Imported message handler from {}".format(handler_module_name))
        except ImportError as e:
            logger.critical(str(e))
            sys.exit(1)

    try:
        handler = handler_module.handler(config)

        if not isinstance(handler, MessageHandler):
            logger.critical("{}.handler is not a subclass of MessageHandler".format(handler_module_name))
            sys.exit(1)
    except (AttributeError, TypeError) as e:
        logger.critical("Cannot initialise message handler from module {}: {}".format(handler_module_name, e))
        sys.exit(1)

    return handler


def determine_interface_configs(config: configparser.ConfigParser):
    """
    Refine the config sections about interfaces. This will expand wildcards, resolve addresses etc.

    :param config: the config parser object
    :return: the list of configured interface names
    """
    interface_names = netifaces.interfaces()

    # Check the interface sections
    for section_name in config.sections():
        parts = section_name.split(' ')

        # Skip non-interface sections
        if parts[0] != 'interface':
            continue

        # Check interface existence
        interface_name = parts[1]
        if interface_name != '*' and interface_name not in interface_names:
            logger.critical("Interface '{}' not found".format(interface_name))
            sys.exit(1)

        section = config[section_name]

        # Add some defaults if necessary
        section.setdefault('multicast', 'no')
        section.setdefault('listen-to-self', 'no')
        if section.getboolean('multicast'):
            # Multicast interfaces need a link-local address
            section.setdefault('link-local-addresses', 'auto')
        else:
            section.setdefault('link-local-addresses', '')
        section.setdefault('global-unicast-addresses', '')

        # make sure these values are booleans
        section.getboolean('multicast')
        section.getboolean('listen-to-self')

        # Make sure that these are 'all', 'auto', or a sequence of addresses
        for option_name in ('link-local-addresses', 'global-unicast-addresses'):
            option_value = section[option_name].lower().strip()

            if option_value not in ('all', 'auto'):
                option_values = set()
                for addr_str in re.split('[,\t ]+', option_value):
                    if not addr_str:
                        # Empty is ok
                        continue

                    try:
                        addr = IPv6Address(addr_str)

                        if option_name == 'link-local-addresses' and not addr.is_link_local:
                            logger.critical("Interface {} option {} must contain "
                                            "link-local addresses".format(interface_name, option_name))
                            sys.exit(1)

                        if option_name == 'global-unicast-addresses' and not (addr.is_global or addr.is_private) \
                                or addr.is_multicast:
                            logger.critical("Interface {} option {} must contain "
                                            "global unicast addresses".format(interface_name, option_name))
                            sys.exit(1)

                        option_values.add(addr)

                    except AddressValueError:
                        logger.critical("Interface {} option {} must contain "
                                        "valid IPv6 addresses".format(interface_name, option_name))
                        sys.exit(1)

                section[option_name] = ' '.join(map(str, option_values))

    # Apply default to unconfigured interfaces
    if config.has_section('interface *'):
        interface_template = config['interface *']

        # Copy from wildcard to other interfaces
        for interface_name in interface_names:
            section_name = 'interface {}'.format(interface_name)
            if config.has_section(section_name):
                # Don't touch it
                pass
            else:
                config.add_section(section_name)
                section = config[section_name]
                for option_name, option_value in interface_template.items():
                    section[option_name] = option_value

        # Forget about the wildcard
        del config['interface *']

    # Expand 'all' and 'auto' and validate the result
    interface_names = [section_name.split(' ')[1] for section_name in config.sections()
                       if section_name.split(' ')[0] == 'interface']
    interface_count = 0
    for interface_name in interface_names:
        section_name = 'interface {}'.format(interface_name)
        section = config[section_name]

        for option_name in ('link-local-addresses', 'global-unicast-addresses'):
            option_value = section[option_name].lower()

            if option_value in ('auto', 'all'):
                logger.debug("Discovering {} on interface {}".format(option_name, interface_name))

                # Get all addresses
                available_addresses_info = netifaces.ifaddresses(interface_name).get(netifaces.AF_INET6, [])
                available_address_strings = [address_info['addr'] for address_info in available_addresses_info]
                available_addresses = [IPv6Address(address.split('%')[0]) for address in available_address_strings]
                available_addresses.sort()

                # Filter on type
                if option_name == 'link-local-addresses':
                    available_addresses = [address for address in available_addresses if address.is_link_local]
                elif option_name == 'global-unicast-addresses':
                    available_addresses = [address for address in available_addresses if not address.is_link_local]

                for address in available_addresses:
                    logger.debug("- Found {}".format(address))

                if option_value == 'all':
                    if available_addresses:
                        logger.debug("= Using all of them")
                    else:
                        logger.debug("= No {} on interface {}: skipping".format(option_name, interface_name))

                elif option_value == 'auto':
                    # Pick the 'best' one if the config says 'auto'
                    # TODO: need to take autoconf/temporary/etc into account once netifaces implements those

                    # First try to find an address with the universal bit set
                    universal_addresses = [address for address in available_addresses if address.packed[8] & 2]
                    if universal_addresses:
                        # Take the lowest universal address
                        available_addresses = [min(universal_addresses)]

                    elif available_addresses:
                        # Take the lowest available address
                        available_addresses = [min(available_addresses)]

                    if available_addresses:
                        logger.debug("= Chose {} as 'best' address".format(available_addresses[0]))
                    else:
                        logger.debug("= No {} on interface {}: skipping".format(option_name, interface_name))

                # Store list of addresses as strings. Yes, this means we probably have to parse them again later but I
                # want to keep the config as clean strings.
                section[option_name] = ' '.join(map(str, available_addresses))

        # Remove interfaces without addresses
        if not section['link-local-addresses'] and not section['global-unicast-addresses']:
            del config[section_name]
            continue

        # Remove loopback interfaces
        if section['global-unicast-addresses']:
            global_addresses = [IPv6Address(address) for address in section['global-unicast-addresses'].split(' ')]
            if any([address.is_loopback for address in global_addresses]):
                # We don't want loopback interfaces
                logger.warning("Not listening on interface {}: it is a loopback interface".format(interface_name))
                del config[section_name]
                continue

        # Check that multicast interfaces have a link-local address
        if section.getboolean('multicast') and not section['link-local-addresses']:
            logger.critical("Interface {} listens for multicast requests "
                            "but has no link-local address to reply from".format(interface_name))
            sys.exit(1)

        # Count this one
        interface_count += 1

    if interface_count == 0:
        logger.critical("This server is not configured to listen on any interfaces")
        sys.exit(1)


def determine_server_duid(config: configparser.ConfigParser):
    """
    Make sure we have a server DUID.

    :param config: The configuration
    """
    # Try to get the server DUID from the configuration
    config_duid = config['server']['duid']
    if config_duid.lower() not in ('', 'auto'):
        config_duid = config_duid.strip()
        try:
            duid = bytes.fromhex(config_duid.strip())
        except ValueError:
            logger.critical("Configured hex DUID contains invalid characters")
            sys.exit(1)

        # Check if we can parse this DUID
        length, duid = DUID.parse(duid, length=len(duid))
        if not isinstance(duid, DUID):
            logger.critical("Configured DUID is invalid")
            sys.exit(1)

        logger.debug("Using server DUID from configuration: {}".format(config_duid))

        config['server']['duid'] = codecs.encode(duid.save(), 'hex').decode('ascii')
        return

    # Use the first interface's MAC address as default
    if config:
        interface_names = [section_name.split(' ')[1] for section_name in config.sections()
                           if section_name.split(' ')[0] == 'interface']
        interface_names.sort()

        for interface_name in interface_names:
            link_addresses = netifaces.ifaddresses(interface_name).get(netifaces.AF_LINK, [])
            link_addresses = [link_address['addr'] for link_address in link_addresses if link_address.get('addr')]
            link_addresses.sort()

            for link_address in link_addresses:
                # Try to decode
                try:
                    ll_addr = bytes.fromhex(link_address.replace(':', ''))

                    duid = LinkLayerDUID(hardware_type=1, link_layer_address=ll_addr).save()

                    logger.debug("Using server DUID based on {} link address: "
                                 "{}".format(interface_name, codecs.encode(duid, 'hex').decode('ascii')))

                    config['server']['duid'] = codecs.encode(duid, 'hex').decode('ascii')
                    return
                except ValueError:
                    # Try the next one
                    pass

    # We didn't find a useful server DUID
    logger.critical("Cannot find a usable DUID")
    sys.exit(1)


def get_sockets(config: configparser.ConfigParser) -> [ListeningSocket]:
    """
    Set up the network sockets.

    :param config: The configuration
    :return: The list of sockets
    """
    logger.debug("Creating sockets")

    mc_address = dhcpkit.ipv6.All_DHCP_Relay_Agents_and_Servers
    port = dhcpkit.ipv6.SERVER_PORT

    # Placeholders for exception message
    interface_name = 'unknown'
    address = 'unknown'

    try:
        sockets = []

        interface_names = [section_name.split(' ')[1] for section_name in config.sections()
                           if section_name.split(' ')[0] == 'interface']
        for interface_name in interface_names:
            section_name = 'interface {}'.format(interface_name)
            section = config[section_name]

            interface_index = socket.if_nametoindex(interface_name)

            first_global = None
            for address_str in section['global-unicast-addresses'].split(' '):
                if not address_str:
                    continue

                address = IPv6Address(address_str)
                logger.debug("- Creating socket for {} on {}".format(address, interface_name))

                sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
                sock.bind((str(address), port))
                sockets.append(ListeningSocket(interface_name, sock))

                if not first_global:
                    first_global = address

            if not first_global and (section['link-local-addresses'] or section.getboolean('multicast')):
                # Get all global addresses because we would like one for link identification
                available_addresses_info = netifaces.ifaddresses(interface_name).get(netifaces.AF_INET6, [])
                available_address_strings = [address_info['addr'] for address_info in available_addresses_info]
                available_addresses = [IPv6Address(address.split('%')[0]) for address in available_address_strings]
                global_addresses = [address for address in available_addresses if not address.is_link_local]
                global_addresses.sort()

                if global_addresses:
                    first_global = global_addresses[0]
                else:
                    # No address known
                    first_global = IPv6Address('::')

            link_local_sockets = []
            for address_str in section['link-local-addresses'].split(' '):
                if not address_str:
                    continue

                address = IPv6Address(address_str)
                logger.debug("- Creating socket for {} on {}".format(address, interface_name))

                sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
                sock.bind((str(address), port, 0, interface_index))
                link_local_sockets.append((address, sock))
                sockets.append(ListeningSocket(interface_name, sock, global_address=first_global))

            if section.getboolean('multicast'):
                address = mc_address
                reply_from = link_local_sockets[0]

                logger.debug("- Creating socket for {} with {} "
                             "as reply-from address on {} ".format(address, reply_from[0], interface_name))

                sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
                sock.bind((address, port, 0, interface_index))

                if section.getboolean('listen-to-self'):
                    sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_LOOP, 1)

                sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_JOIN_GROUP,
                                pack('16sI', IPv6Address('ff02::1:2').packed, interface_index))

                sockets.append(ListeningSocket(interface_name, sock, reply_from[1], global_address=first_global))

    except OSError as e:
        logger.critical("Cannot create socket for address {} on interface {}: {}".format(address, interface_name,
                                                                                         e.strerror))
        sys.exit(1)
    except ListeningSocketError as e:
        logger.critical(str(e))
        sys.exit(1)

    return sockets


def drop_privileges(uid_name: str or int, gid_name: str or int or None):
    """
    Drop root privileges and change to something more safe.

    :param uid_name: The UID to drop to
    :param gid_name: The GID to drop to
    """
    if os.getuid() != 0:
        logger.warning("Not running as root: cannot change uid/gid to {}/{}".format(uid_name, gid_name))
        return

    # Get the uid/gid from the name
    try:
        try:
            # Try to use it as an integer
            running_uid = int(uid_name)
        except ValueError:
            # Try to use it as a name
            running_uid = pwd.getpwnam(uid_name).pw_uid
    except KeyError:
        logger.critical("User {} does not exist".format(uid_name))
        sys.exit(1)

    if gid_name:
        try:
            try:
                # Try to use it as an integer
                running_gid = int(gid_name)
            except ValueError:
                # Try to use it as a name
                running_gid = grp.getgrnam(gid_name).gr_gid
        except KeyError:
            logger.critical("Group {} does not exist".format(gid_name))
            sys.exit(1)
    else:
        running_gid = pwd.getpwnam(uid_name).pw_gid

    # Remove group privileges
    os.setgroups([])

    # Try setting the new uid/gid
    os.setgid(running_gid)
    os.setuid(running_uid)

    # Ensure a very conservative umask
    os.umask(0o077)

    # Resolve names
    try:
        uid_name = pwd.getpwuid(running_uid).pw_name
    except KeyError:
        uid_name = str(running_uid)

    try:
        gid_name = grp.getgrgid(running_gid).gr_name
    except KeyError:
        gid_name = str(running_gid)

    logger.debug("Dropped privileges to {}/{}".format(uid_name, gid_name))


def create_handler_callback(listening_socket: ListeningSocket) -> types.FunctionType:
    """
    Create a callback for the handler method that still knows the listening socket and the sender

    :param listening_socket: The listening socket to remember
    :return: A callback function with the listening socket and sender enclosed
    :rtype: (concurrent.futures.Future) -> None
    """

    def callback(future: concurrent.futures.Future):
        """
        A callback that handles the result of a handler

        :param future: The future object with the completed result
        """
        try:
            # Get the result
            reply = future.result()

            if reply is None:
                # No reply: we're done with this request
                return

            if not isinstance(reply, RelayReplyMessage):
                logger.error("Handler returned invalid result, not sending a reply")
                return

            try:
                listening_socket.send_reply(reply)
            except ValueError as e:
                logger.error("Handler returned invalid message: {}".format(e))
                return

        except concurrent.futures.CancelledError:
            pass

        except Exception as e:
            # Catch-all exception handler
            logger.exception("Caught unexpected exception {!r}".format(e))

    return callback


def main() -> int:
    """
    The main program loop

    :return: The program exit code
    """
    args = handle_args()
    config = load_config(args.config)
    set_up_logger(config, args.verbosity)

    logger.info("Starting Python DHCPv6 server v{}".format(dhcpkit.__version__))

    determine_interface_configs(config)
    determine_server_duid(config)

    if args.show_config:
        config.write(sys.stdout)
        sys.exit(0)

    sockets = get_sockets(config)
    drop_privileges(config['server']['user'], config['server']['group'])

    handler = get_handler(config)

    sel = selectors.DefaultSelector()
    for sock in sockets:
        sel.register(sock, selectors.EVENT_READ)

    # Convert signals to messages on a pipe
    signal_r, signal_w = os.pipe()
    flags = fcntl.fcntl(signal_w, fcntl.F_GETFL, 0)
    flags = flags | os.O_NONBLOCK
    fcntl.fcntl(signal_w, fcntl.F_SETFL, flags)
    signal.set_wakeup_fd(signal_w)
    sel.register(signal_r, selectors.EVENT_READ)

    # Ignore normal signal handling
    signal.signal(signal.SIGINT, lambda signum, frame: None)
    signal.signal(signal.SIGTERM, lambda signum, frame: None)
    signal.signal(signal.SIGHUP, lambda signum, frame: None)

    # Excessive exception catcher
    exception_history = []

    logger.info("Python DHCPv6 server is ready to handle requests")

    exception_window = config['server'].getfloat('exception-window')
    max_exceptions = config['server'].getint('max-exceptions')
    workers = max(1, config['server'].getint('threads'))
    with concurrent.futures.ThreadPoolExecutor(max_workers=workers) as executor:
        stopping = False
        while not stopping:
            # noinspection PyBroadException
            try:
                events = sel.select()
                for key, mask in events:
                    # Handle signal notifications
                    if key.fileobj == signal_r:
                        signal_nr = os.read(signal_r, 1)
                        if signal_nr[0] in (signal.SIGHUP,):
                            # SIGHUP tells the handler to reload
                            # We might even re-parse the config in a later implementation
                            handler.reload(config)
                        elif signal_nr[0] in (signal.SIGINT, signal.SIGTERM):
                            logger.debug("Received termination request")

                            stopping = True
                            break

                        # Unknown signal: ignore
                        continue
                    elif isinstance(key.fileobj, ListeningSocket):
                        try:
                            msg_in = key.fileobj.recv_request()
                        except InvalidPacketError as e:
                            logging.warning("Invalid message from {}: {}".format(e.sender[0], str(e)))
                            continue
                        except ValueError as e:
                            logging.warning("Invalid incoming message: {}".format(str(e)))
                            continue

                        # Submit this request to the worker pool
                        received_over_multicast = key.fileobj.listen_address.is_multicast
                        future = executor.submit(handler.handle, msg_in, received_over_multicast)

                        # Create the callback
                        callback = create_handler_callback(key.fileobj)
                        future.add_done_callback(callback)

            except Exception as e:
                # Catch-all exception handler
                logger.exception("Caught unexpected exception {!r}".format(e))

                now = time.monotonic()

                # Add new exception time to the history
                exception_history.append(now)

                # Remove exceptions outside the window from the history
                cutoff = now - exception_window
                while exception_history and exception_history[0] < cutoff:
                    exception_history.pop(0)

                # Did we receive too many exceptions shortly after each other?
                if len(exception_history) > max_exceptions:
                    logger.critical("Received more than {} exceptions in {} seconds, exiting".format(max_exceptions,
                                                                                                     exception_window))
                    stopping = True

    logger.info("Shutting down Python DHCPv6 server v{}".format(dhcpkit.__version__))

    return 0


def run() -> int:
    """
    Run the main program and handle exceptions

    :return: The program exit code
    """
    try:
        return main()
    except configparser.Error as e:
        logger.critical("Configuration error: {}".format(e))
        sys.exit(1)


if __name__ == '__main__':
    # Run the server
    sys.exit(run())



"""
The option registry
"""
from dhcpkit.registry import Registry


class OptionRegistry(Registry):
    """
    Registry for DHCPKit IPv6 Options
    """
    entry_point = 'dhcpkit.ipv6.options'


# Instantiate the option registry
option_registry = OptionRegistry()



"""
Option handlers for the DNS options defined in dhcpkit.ipv6.extensions.sntp
"""

import configparser
from ipaddress import IPv6Address

from dhcpkit.ipv6.extensions.sntp import SNTPServersOption
from dhcpkit.ipv6.option_handlers import SimpleOptionHandler, OptionHandler


class SNTPServersOptionHandler(SimpleOptionHandler):
    """
    Handler for putting SNTPServersOptions in responses
    """

    def __init__(self, sntp_servers: [IPv6Address]):
        option = SNTPServersOption(sntp_servers=sntp_servers)
        option.validate()

        super().__init__(option)

    @classmethod
    def from_config(cls, section: configparser.SectionProxy, option_handler_id: str = None) -> OptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :param section: The configuration section
        :param option_handler_id: Optional extra identifier
        :return: A handler object
        :rtype: OptionHandler
        """
        addresses = []
        for name, value in section.items():
            # Strip numbers from the end, this can be used to supply the same option multiple times
            name = name.rstrip('0123456789-')

            if name != 'server-address':
                continue

            addresses.append(IPv6Address(value))

        return cls(addresses)



"""
Option handler that implements rapid-commit on the server.
"""

from dhcpkit.ipv6.extensions.prefix_delegation import IAPDOption, STATUS_NOPREFIXAVAIL
from dhcpkit.ipv6.messages import SolicitMessage, AdvertiseMessage, ReplyMessage
from dhcpkit.ipv6.option_handlers import OptionHandler
from dhcpkit.ipv6.options import RapidCommitOption, IANAOption, IATAOption, StatusCodeOption, STATUS_NOADDRSAVAIL
from dhcpkit.ipv6.transaction_bundle import TransactionBundle


class RapidCommitOptionHandler(OptionHandler):
    """
    Upgrade AdvertiseMessage to ReplyMessage when client asks for rapid-commit
    """

    def __init__(self, rapid_commit_rejections: bool):
        super().__init__()

        self.rapid_commit_rejections = rapid_commit_rejections
        """
        Do rapid-commit when an IA_NA, IA_TA or IA_PD request gets refused. We have seen at lease one
        vice (Fritz!Box) that gets confused when a rapid-commit message tells it there are no addresses
        available. Turning this setting off works around that problem by not doing a rapid-commit when
        something gets refused.
        """

    def handle(self, bundle: TransactionBundle):
        """
        Don't do anything, all the processing happens in :meth:`post`.

        :param bundle: The transaction bundle
        """

    def post(self, bundle: TransactionBundle):
        """
        Upgrade the response from a AdvertiseMessage to a ReplyMessage if appropriate
        :param bundle: The transaction bundle
        """
        # Does this transaction even allow rapid commit?
        if not bundle.allow_rapid_commit:
            return

        # We only look for SolicitMessages that have a RapidCommitOption
        if not isinstance(bundle.request, SolicitMessage) or not bundle.request.get_option_of_type(RapidCommitOption):
            return

        # And only if the current response is an AdvertiseMessage
        if not isinstance(bundle.response, AdvertiseMessage):
            return

        # Ok, this looks promising, do extra checks if requested
        if not self.rapid_commit_rejections:
            # Ok, we don't want to rapid-commit rejections. Check for them.
            if bundle.get_unanswered_ia_options() or bundle.get_unanswered_iapd_options():
                # Unhandled options. We are post-processing, so they are not going to be answered anymore
                return

            # Did we already refuse anything?
            ia_options = [option for option in bundle.response.options if isinstance(option, (IANAOption, IATAOption))]
            for option in ia_options:
                status = option.get_option_of_type(StatusCodeOption)
                if status and status.status_code == STATUS_NOADDRSAVAIL:
                    # Refusal: don't do anything
                    return

            iapd_options = [option for option in bundle.response.options if isinstance(option, IAPDOption)]
            for option in iapd_options:
                status = option.get_option_of_type(StatusCodeOption)
                if status and status.status_code == STATUS_NOPREFIXAVAIL:
                    # Refusal: don't do anything
                    return

        # It seems the request and response qualify: upgrade to ReplyMessage
        bundle.response = ReplyMessage(bundle.response.transaction_id, [RapidCommitOption()] + bundle.response.options)



"""
The base class :class:`ProtocolElement` provides the basic structure for each element of the DHCP protocol. This base
class provides several functions:

- Parsing:
    Each subclass can parse a stream of bytes from a protocol packet and construct an instance that contains all the
    data from the byte stream as properties.

- Identification:
    Each category of ProtocolElement can determine which subclass is the most specific implementation for the data
    being parsed. For example when letting the Message class :func:`parse <Message.parse>` a message it will look at
    the message type code in the byte steam and determine which specific subclass should parse the data (i.e.
    SolicitMessage, RequestMessage, ReplyMessage etc). Each category of ProtocolElement has its own registry that keeps
    track of which type code corresponds to which subclass.

- Saving:
    Each instance can save its contents to a stream of bytes as required by
    the protocol.

- Validation:
    Each element can validate if its contents are valid. As protocol elements
    often contain other protocol elements (a message has options, an option
    might have sub-options etc) there are standard tools for defining which
    protocol element may contain which other protocol elements and optionally
    define a minimum and maximum occurrence. Some elements may not occur more
    than once, some elements must occur at least once, etc.

- Representation:
    The default implementation provides __str__ and __repr__ methods so that
    protocol elements can be printed for debugging and represented as a
    parsable Python string.
"""

from abc import abstractmethod, ABCMeta
from collections import ChainMap
from inspect import Parameter
import inspect
import collections

infinite = 2 ** 31 - 1


class AutoMayContainTree(ABCMeta):
    """
    Meta-class that automatically creates a _may_contain class property that is a ChainMap that links all
    parent _may_contain class properties.
    """

    def __new__(mcs, name, bases, namespace):
        cls = super().__new__(mcs, name, bases, namespace)

        # Get all the ChainMaps from the parents
        parent_may_contains = [getattr(base, '_may_contain') for base in bases
                               if isinstance(getattr(base, '_may_contain', None), ChainMap)]

        # And create our local one with those as lookup targets
        cls._may_contain = ChainMap({}, *parent_may_contains)

        return cls


class ProtocolElement(metaclass=AutoMayContainTree):
    """
    A StructuredElement is a specific kind of class that represents a protocol message or option. Structured elements
    have the following extra requirements:

    - The constructor parameters and the internal state properties must be identical
      So if an object has a property `timeout` which is an integer then the constructor must accept a named parameter
      called `timeout` which is stored in that property. The constructor must have appropriate default values if
      possible. Empty objects, lists, dictionaries etc are represented by a default value of None.
    - The full internal state of the object must be loadable from a bytes object with the :func:`load_from` method
    - The full internal state of the object must be storable as a bytes object with the :func:`save` method
    """

    # This will be set by the meta-class
    _may_contain = None

    def validate(self):
        """
        Subclasses may overwrite this method to validate their state. Subclasses are expected to raise a ValueError
        if validation fails.
        """
        pass

    def validate_contains(self, elements: [object]):
        """
        Utility method that subclasses can use in their validate method for verifying that all sub-elements are allowed
        to be contained in this element. Will raise ValueError if validation fails.

        :param elements: The list of sub-elements
        """
        # Count occurrence
        occurrence_counters = collections.Counter()
        for element in elements:
            element_class = self.get_element_class(element)
            if element_class is None:
                raise ValueError("{} cannot contain {}".format(self.__class__.__name__, element.__class__.__name__))

            # Count its occurrence
            occurrence_counters[element_class] += 1

        # Check max occurrence
        for element_class, (min_occurrence, max_occurrence) in self._may_contain.items():
            count = occurrence_counters[element_class]
            if count > max_occurrence:
                if max_occurrence == 1:
                    raise ValueError("{} may only contain 1 {}".format(self.__class__.__name__, element_class.__name__))
                else:
                    raise ValueError("{} may only contain {} {}s".format(self.__class__.__name__, max_occurrence,
                                                                         element_class.__name__))
            elif count < min_occurrence:
                if min_occurrence == 1:
                    raise ValueError("{} must contain at least 1 {}".format(self.__class__.__name__,
                                                                            element_class.__name__))
                else:
                    raise ValueError("{} must contain at least {} {}s".format(self.__class__.__name__, max_occurrence,
                                                                              element_class.__name__))

    @classmethod
    @abstractmethod
    def determine_class(cls, buffer: bytes, offset: int=0) -> type:
        """
        Return the appropriate class to parse this element with.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :return: The best known class for this data
        """

    @classmethod
    def parse(cls, buffer: bytes, offset: int=0, length: int=None) -> (int, type):
        """
        Constructor for a new element of which the state is automatically loaded from the given buffer. Both the number
        of bytes used from the buffer and the instantiated element are returned. The class of the returned element may
        be a subclass of the current class if the parser can determine that the data in the buffer contains a subtype.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer and the resulting element
        """
        element_class = cls.determine_class(buffer, offset=offset)
        element = element_class()
        length = element.load_from(buffer, offset=offset, length=length)
        return length, element

    @abstractmethod
    def load_from(self, buffer: bytes, offset: int=0, length: int=None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """

    @abstractmethod
    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """

    def __eq__(self, other: object) -> bool:
        """
        Compare this object to another object. The result will be True if they are of the same class and if the
        properties have equal values and False otherwise.

        :param other: The other object
        :return: Whether this object is equal to the other one
        """
        # Use strict comparison, one being a subclass of the other is not good enough
        if type(self) is not type(other):
            return NotImplemented

        # Get the signature of the __init__ method to find the properties we need to compare
        # This is why the object properties and __init__ parameters need to match, besides it being good practice for
        # an object that represents a protocol element anyway...
        signature = inspect.signature(self.__init__)

        # Compare the discovered properties
        for parameter in signature.parameters.values():
            # Skip any potential *args and **kwargs in the method signature
            if parameter.kind in (Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD):
                continue

            if getattr(self, parameter.name) != getattr(other, parameter.name):
                return False

        # Amazing, all properties seem equal
        return True

    def __repr__(self):
        """
        Return a machine-readable representation of this protocol element.

        :return: Parsable representation of this protocol element
        """
        # Get the signature of the __init__ method to find the properties we need to compare
        # This is why the object properties and __init__ parameters need to match, besides it being good practice for
        # an object that represents a protocol element anyway...
        signature = inspect.signature(self.__init__)

        # Create a list of string with "parameter=value" for each parameter of __init__
        options_repr = ['{}={}'.format(parameter.name, repr(getattr(self, parameter.name)))
                        for parameter in signature.parameters.values()
                        if parameter.kind not in (Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD)]

        # And construct a constructor call to show
        return '{}({})'.format(self.__class__.__name__, ', '.join(options_repr))

    def __str__(self):
        """
        Return a human-readable and indented representation of this protocol element.

        :return: Readable representation of this protocol element
        """
        # Use the same strategy as __repr__ but do nice indenting etc.
        signature = inspect.signature(self.__init__)

        parameter_names = [parameter.name
                           for parameter in signature.parameters.values()
                           if parameter.kind not in (Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD)]

        if len(parameter_names) == 0:
            # No parameters: inline
            return '{}()'.format(self.__class__.__name__)
        elif len(parameter_names) == 1:
            # One parameter: inline unless the parameter has a multi-line string output
            parameter_name = parameter_names[0]
            attr_value = getattr(self, parameter_name)
            lines = str(attr_value).split('\n')

            output = '{}('.format(self.__class__.__name__, parameter_name)

            if len(lines) == 1:
                output += '{}={}'.format(parameter_name, lines[0])
            else:
                output += '\n  {}='.format(parameter_name)
                output += '{}\n'.format(lines[0])
                for line in lines[1:-1]:
                    output += '  {}\n'.format(line)
                output += '  {}\n'.format(lines[-1])

            output += ')'
            return output

        # Multiple parameters are shown one parameter per line
        output = '{}(\n'.format(self.__class__.__name__)
        for parameter_name in parameter_names:
            attr_value = getattr(self, parameter_name)

            if attr_value and isinstance(attr_value, str):
                # Show strings with repr()
                attr_value = repr(getattr(self, parameter_name))

            if attr_value and isinstance(attr_value, list):
                # Parameters containing lists show the list content indented
                output += '  {}=[\n'.format(parameter_name)
                for element in attr_value:
                    lines = str(element).split('\n')
                    for line in lines[:-1]:
                        output += '    {}\n'.format(line)
                    output += '    {},\n'.format(lines[-1])
                output += '  ],\n'
            else:
                # Multi-line content is shown indented
                output += '  {}='.format(parameter_name)
                lines = str(attr_value).split('\n')
                if len(lines) == 1:
                    output += '{},\n'.format(lines[0])
                else:
                    output += '{}\n'.format(lines[0])
                    for line in lines[1:-1]:
                        output += '  {}\n'.format(line)
                    output += '  {},\n'.format(lines[-1])

        output += ')'

        return output

    @classmethod
    def add_may_contain(cls, klass: type, min_occurrence: int=0, max_occurrence: int=infinite):
        """
        Add the given class to the list of permitted sub-element classes, optionally with a minimum and maximum
        occurrence count.

        :param klass: The class to add
        :param min_occurrence: Minimum occurrence for validation
        :param max_occurrence: Maximum occurrence for validation
        """
        cls._may_contain[klass] = (min_occurrence, max_occurrence)

    @classmethod
    def may_contain(cls, element: object) -> bool:
        """
        Shortcut-method to verify that objects of this class may contain element

        :param element: Sub-element to verify
        :return: Whether this class may contain element or not
        """
        return cls.get_element_class(element) is not None

    @classmethod
    def get_element_class(cls, element: object) -> type:
        """
        Get the class this element is classified as, for occurrence counting.

        :param element: Some element
        :return: The class it classifies as
        """
        # This class has its own list of what it may contain: check it
        for klass, (min_occurrence, max_occurrence) in cls._may_contain.items():
            if max_occurrence < 1:
                # May not contain this, stop looking
                return None

            if inspect.isclass(element):
                if issubclass(element, klass):
                    return klass
            elif isinstance(element, klass):
                return klass



"""
Option handlers that cleans up unanswered requests
"""
import configparser
import logging

from dhcpkit.ipv6.exceptions import CannotRespondError
from dhcpkit.ipv6.extensions.prefix_delegation import IAPDOption, STATUS_NOPREFIXAVAIL, IAPrefixOption
from dhcpkit.ipv6.messages import SolicitMessage, RequestMessage, RenewMessage, RebindMessage, ReleaseMessage, \
    DeclineMessage, ConfirmMessage
from dhcpkit.ipv6.option_handlers import OptionHandler
from dhcpkit.ipv6.option_handlers.utils import force_status
from dhcpkit.ipv6.options import StatusCodeOption, STATUS_NOADDRSAVAIL, STATUS_NOBINDING, \
    STATUS_NOTONLINK, IAAddressOption
from dhcpkit.ipv6.transaction_bundle import TransactionBundle

logger = logging.getLogger(__name__)


class UnansweredIAOptionHandler(OptionHandler):
    """
    A handler that answers to all unanswered IANAOptions and IATAOptions

    :param authoritative: Whether this handler is authorised to tell clients to stop using prefixes
    """

    def __init__(self, authoritative: bool=True):
        self.authoritative = authoritative

    @classmethod
    def from_config(cls, section: configparser.SectionProxy, option_handler_id: str=None) -> OptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :param section: The configuration section
        :param option_handler_id: Optional extra identifier
        :return: A handler object
        :rtype: OptionHandler
        """
        authoritative = section.getboolean('authoritative', False)
        return cls(authoritative)

    def handle(self, bundle: TransactionBundle):
        """
        Make sure that every :class:`.IANAOption` and :class:`.IATAOption` is answered.

        :param bundle: The transaction bundle
        """
        for option in bundle.get_unanswered_ia_options():
            ia_class = type(option)

            if isinstance(bundle.request, (SolicitMessage, RequestMessage)):
                # If the server will not assign any addresses to any IAs in a subsequent Request from the client, the
                # server MUST send an Advertise message to the client that includes only a Status Code option with code
                # NoAddrsAvail and a status message for the user
                #
                # We do the same for unanswered requests
                bundle.response.options.append(ia_class(option.iaid, options=[
                    StatusCodeOption(STATUS_NOADDRSAVAIL, "No addresses available")
                ]))

            elif isinstance(bundle.request, ConfirmMessage):
                # When the server receives a Confirm message, the server determines whether the addresses in the
                # Confirm message are appropriate for the link to which the client is attached.  If all of the
                # addresses in the Confirm message pass this test, the server returns a status of Success.  If any of
                # the addresses do not pass this test, the server returns a status of NotOnLink.  If the server is
                # unable to perform this test (for example, the server does not have information about prefixes on the
                # link to which the client is connected), or there were no addresses in any of the IAs sent by the
                # client, the server MUST NOT send a reply to the client.
                #
                # The "there were no addresses in any of the IAs sent by the client" check is done by the message
                # handler.
                if not self.authoritative:
                    raise CannotRespondError

                addresses = ', '.join([str(suboption.address)
                                       for suboption in option.get_options_of_type(IAAddressOption)])
                logger.warning("No handler confirmed {} for {}: "
                               "sending NotOnLink status".format(addresses, bundle.get_link_address()))

                force_status(bundle.response.options,
                             StatusCodeOption(STATUS_NOTONLINK, "Those addresses are not appropriate on this link"))

            elif isinstance(bundle.request, RenewMessage):
                # If the server cannot find a client entry for the IA the server returns the IA containing no addresses
                # with a Status Code option set to NoBinding in the Reply message.
                #
                # If the server finds that any of the addresses are not appropriate for the link to which the client is
                # attached, the server returns the address to the client with lifetimes of 0.
                addresses = ', '.join([str(suboption.address)
                                       for suboption in option.get_options_of_type(IAAddressOption)])

                if self.authoritative:
                    logger.warning("No handler renewed {} for {}: "
                                   "withdrawing addresses".format(addresses, bundle.get_link_address()))

                    reply_suboptions = []
                    for suboption in option.get_options_of_type(IAAddressOption):
                        reply_suboptions.append(IAAddressOption(suboption.address,
                                                                preferred_lifetime=0, valid_lifetime=0))

                    bundle.response.options.append(ia_class(option.iaid, options=reply_suboptions))
                else:
                    logger.warning("No handler renewed {} for {}: "
                                   "sending NoBinding status".format(addresses, bundle.get_link_address()))

                    bundle.response.options.append(ia_class(option.iaid, options=[
                        StatusCodeOption(STATUS_NOBINDING, "No addresses assigned to you")
                    ]))

            elif isinstance(bundle.request, RebindMessage):
                # If the server cannot find a client entry for the IA and the server determines that the addresses in
                # the IA are not appropriate for the link to which the client's interface is attached according to the
                # server's explicit configuration information, the server MAY send a Reply message to the client
                # containing the client's IA, with the lifetimes for the addresses in the IA set to zero.  This Reply
                # constitutes an explicit notification to the client that the addresses in the IA are no longer valid.
                # In this situation, if the server does not send a Reply message it silently discards the Rebind
                # message.
                #
                # If the server finds that any of the addresses are no longer appropriate for the link to which the
                # client is attached, the server returns the address to the client with lifetimes of 0.
                if not self.authoritative:
                    raise CannotRespondError

                addresses = ', '.join([str(suboption.address)
                                       for suboption in option.get_options_of_type(IAAddressOption)])
                logger.warning("No handler answered rebind of {} for {}: "
                               "withdrawing addresses".format(addresses, bundle.get_link_address()))

                reply_suboptions = []
                for suboption in option.get_options_of_type(IAAddressOption):
                    reply_suboptions.append(IAAddressOption(suboption.address, preferred_lifetime=0, valid_lifetime=0))

                bundle.response.options.append(ia_class(option.iaid, options=reply_suboptions))

            elif isinstance(bundle.request, DeclineMessage):
                # For each IA in the Decline message for which the server has no binding information, the server adds
                # an IA option using the IAID from the Release message and includes a Status Code option with the value
                # NoBinding in the IA option.  No other options are included in the IA option.
                bundle.response.options.append(ia_class(option.iaid, options=[
                    StatusCodeOption(STATUS_NOBINDING, "No addresses assigned to you")
                ]))

            elif isinstance(bundle.request, ReleaseMessage):
                # For each IA in the Release message for which the server has no binding information, the server adds an
                # IA option using the IAID from the Release message, and includes a Status Code option with the value
                # NoBinding in the IA option.  No other options are included in the IA option.
                bundle.response.options.append(ia_class(option.iaid, options=[
                    StatusCodeOption(STATUS_NOBINDING, "No addresses assigned to you")
                ]))


class UnansweredIAPDOptionHandler(OptionHandler):
    """
    A handler that answers to all unanswered IAPDOptions

    :param authoritative: Whether this handler is authorised to tell clients to stop using prefixes
    """

    def __init__(self, authoritative: bool=True):
        self.authoritative = authoritative

    @classmethod
    def from_config(cls, section: configparser.SectionProxy, option_handler_id: str=None) -> OptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :param section: The configuration section
        :param option_handler_id: Optional extra identifier
        :return: A handler object
        :rtype: OptionHandler
        """
        authoritative = section.getboolean('authoritative', False)
        return cls(authoritative)

    def handle(self, bundle: TransactionBundle):
        """
        Make sure that every :class:`.IAPDOption` is answered.

        :param bundle: The transaction bundle
        """
        for option in bundle.get_unanswered_iapd_options():
            if isinstance(bundle.request, (SolicitMessage, RequestMessage)):
                # If the delegating router will not assign any prefixes to any IA_PDs in a subsequent Request from the
                # requesting router, the delegating router MUST send an Advertise message to the requesting router that
                # includes the IA_PD with no prefixes in the IA_PD and a Status Code option in the IA_PD containing
                # status code NoPrefixAvail and a status message for the user
                #
                # We do the same for unanswered requests
                bundle.response.options.append(IAPDOption(option.iaid, options=[
                    StatusCodeOption(STATUS_NOPREFIXAVAIL, "No prefixes available")
                ]))

            elif isinstance(bundle.request, RenewMessage):
                # Renew message: If the delegating router cannot find a binding for the requesting router's IA_PD the
                # delegating router returns the IA_PD containing no prefixes with a Status Code option set to
                # NoBinding in the Reply message.

                prefixes = ', '.join([str(suboption.prefix)
                                      for suboption in option.get_options_of_type(IAPrefixOption)])
                logger.warning("No handler renewed {} for {}: "
                               "sending NoBinding status".format(prefixes, bundle.get_link_address()))

                bundle.response.options.append(IAPDOption(option.iaid, options=[
                    StatusCodeOption(STATUS_NOBINDING, "No prefixes assigned to you")
                ]))

            elif isinstance(bundle.request, RebindMessage):
                # Rebind message: If the delegating router cannot find a binding for the requesting router's IA_PD and
                # the delegating router determines that the prefixes in the IA_PD are not appropriate for the link to
                # which the requesting router's interface is attached according to the delegating routers explicit
                # configuration, the delegating router MAY send a Reply message to the requesting router containing
                # the IA_PD with the lifetimes of the prefixes in the IA_PD set to zero.  This Reply constitutes an
                # explicit notification to the requesting router that the prefixes in the IA_PD are no longer valid.
                #
                # If the delegating router is unable to determine if the prefix is not appropriate for the link, the
                # Rebind message is discarded.
                #
                # The authoritative flag indicates whether this option may claim whether it is able to determine if a
                # prefix is appropriate for the link.
                if not self.authoritative:
                    raise CannotRespondError

                prefixes = ', '.join([str(suboption.prefix)
                                      for suboption in option.get_options_of_type(IAPrefixOption)])
                logger.warning("No handler answered rebind of {} for {}: "
                               "withdrawing prefixes".format(prefixes, bundle.get_link_address()))

                reply_suboptions = []
                for suboption in option.get_options_of_type(IAPrefixOption):
                    reply_suboptions.append(IAPrefixOption(suboption.prefix, preferred_lifetime=0, valid_lifetime=0))

                bundle.response.options.append(IAPDOption(option.iaid, options=reply_suboptions))

            elif isinstance(bundle.request, ReleaseMessage):
                # For each IA in the Release message for which the server has no binding information, the server adds an
                # IA option using the IAID from the Release message, and includes a Status Code option with the value
                # NoBinding in the IA option.  No other options are included in the IA option.
                bundle.response.options.append(IAPDOption(option.iaid, options=[
                    StatusCodeOption(STATUS_NOBINDING, "No prefixes assigned to you")
                ]))



"""
Test the OptionRequestOption implementation
"""
import unittest

from dhcpkit.ipv6.options import OptionRequestOption, OPTION_IA_NA, OPTION_IA_TA
from tests.ipv6.options import test_option


class OptionRequestOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('0006000400030004')
        self.option_object = OptionRequestOption(requested_options=[OPTION_IA_NA, OPTION_IA_TA])
        self.parse_option()

    def test_validate_requested_options(self):
        self.option.requested_options = [OPTION_IA_NA, 0, OPTION_IA_TA]
        self.option.validate()

        self.option.requested_options = [OPTION_IA_NA, -1, OPTION_IA_TA]
        with self.assertRaisesRegex(ValueError, 'unsigned 16 bit integer'):
            self.option.validate()

        self.option.requested_options = [OPTION_IA_NA, 65535, OPTION_IA_TA]
        self.option.validate()

        self.option.requested_options = [OPTION_IA_NA, 65536, OPTION_IA_TA]
        with self.assertRaisesRegex(ValueError, 'unsigned 16 bit integer'):
            self.option.validate()

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'Invalid option length'):
            OptionRequestOption.parse(bytes.fromhex('0006000300030004'))


if __name__ == '__main__':
    unittest.main()
